//----------------------------------------------------------------------------------------------

// CopyJob.cp

//----------------------------------------------------------------------------------------------

/*

Please note the lack of the usual copyright notices.  Please feel free to use this as the basis
for your programming.

*/

#include "CopyJob.h"
#include "CopyUtils.h"
#include "OpenClipDefines.h"

//----------------------------------------------------------------------------------------------

CopyJob::CopyJob(DWindow *TheWndw)
  : mTheWndw(TheWndw),
    mPgSzY(),
    mScale(1.0),
    mMode(kForFlatFile),  // kForFlatFile,kHandleLayers
    mHandleGroups(false),
    mClearPasteboard(false)
{
  CDPoint PgSz = TheWndw->pageSize;
  mPgSzY = PgSz.y;
};


void CopyJob::SetCopyMode(OpenClipCopyMode TheMode)
{
  mMode = TheMode;
};


void CopyJob::SetHandleGroups(bool SetIt)
{
  mHandleGroups = SetIt;
};


void CopyJob::SetClearPasteboard(bool SetIt)
{
  mClearPasteboard = SetIt;
};


void CopyJob::HandleZHeight(const PDObject &TheObj,NSMutableDictionary *ThisObj) const
{
  // SetObjectZ(12345.0,ThisObj);  // to test
  
  // IMPORTANT.  If your program stores a Z height in 'world coordinates' where everything is
  // 1:1 scale, you will need to scale the Z height here.  In fact, the optional topo Z height
  // in PowerCADD is stored at 1:1 scale, so when we to a 3D copy, we do this:
  
  //PDObject TempObj = TheObj;
  
  //double Z;
  //if (HasAZElevation(&TempObj, Z)) {
  //  Z /= mScale;
  //  SetObjectZ(Z,ThisObj);
  //};
  
};


CDPoint CopyJob::CorrectPoint(CDPoint ThePt) const
{
  ThePt.y = mPgSzY - ThePt.y;  // PowerCADD has the origin at the top left of the window
                               // with positive Y values going down
                               
  // IMPORTANT, in PowerCADD all drawing coordinates are the 'on paper' size same as in a PDF
  // and using the drawing scale, PowerCADD 'tells' the user the dimensions are in whatever
  // scale the user has set up.  
  // Most CAD programs deal in 'world coordinates' where everything is 1:1 scale, so if your
  // program is like that, you'll need to scale the coordinates of the point here.
  
  // So it would be
  // ThePt.x /= mScale;
  // ThePt.y /= mScale;
  
  return ThePt; 
};


void CopyJob::SetScale(double TheScale,NSMutableDictionary *TheDict) const
{
  NSNumber *nTheScale = [NSNumber numberWithDouble:TheScale];
  [TheDict setObject:nTheScale forKey:OpenClip_DrawingScale];
};


void CopyJob::SetGlobalOrigin(double X,double Y,NSMutableDictionary *TheDict) const
{
  NSNumber *nX = [NSNumber numberWithDouble:X];
  NSNumber *nY = [NSNumber numberWithDouble:Y];
  
  NSMutableDictionary *PtDict	= [NSMutableDictionary dictionary];
  [PtDict setObject:nX forKey:OpenClip_X];
  [PtDict setObject:nY forKey:OpenClip_Y];
  
  [TheDict setObject:PtDict forKey:OpenClip_GlobalOrigin];
};


void CopyJob::SetInsertionPoint(double X,double Y,NSMutableDictionary *TheDict) const
{
  NSNumber *nX = [NSNumber numberWithDouble:X];
  NSNumber *nY = [NSNumber numberWithDouble:Y];
  
  NSMutableDictionary *PtDict	= [NSMutableDictionary dictionary];
  [PtDict setObject:nX forKey:OpenClip_X];
  [PtDict setObject:nY forKey:OpenClip_Y];
  
  [TheDict setObject:PtDict forKey:OpenClip_InsertionPoint];
};


void CopyJob::SetObjectType(NSString *TheType,NSMutableDictionary *TheObj) const
{
  [TheObj setObject:TheType forKey:OpenClip_ObjectType];
};


void CopyJob::SetLayerName(CStr255 TheName,NSMutableDictionary *TheNSLayer) const
{
  //NSString *TheStr = [NSString stringWithUTF8String_:TheName.c_str()];
  NSString *TheStr = [NSString stringWithFormat:@"%s",TheName.c_str()];
  [TheNSLayer setObject:TheStr forKey:OpenClip_LayerName];
};


void CopyJob::SetLayerColor(RGBColor TheColor,NSMutableDictionary *TheNSLayer) const
{
  NSMutableDictionary *TheLayerColor = RGBADictionary(TheColor,1.0);
  [TheNSLayer setObject:TheLayerColor forKey:OpenClip_LayerColor];
};


void CopyJob::SetLayerScale(double TheScale,NSMutableDictionary *TheNSLayer) const
{
  NSNumber *nTheScale = [NSNumber numberWithDouble:TheScale];
  [TheNSLayer setObject:nTheScale forKey:OpenClip_LayerScale];
};


void CopyJob::AddXYPoint(double X,double Y,NSMutableArray *Points) const
{
  NSNumber *nX = [NSNumber numberWithDouble:X];
  NSNumber *nY = [NSNumber numberWithDouble:Y];
  
  NSMutableDictionary *PtDict	= [NSMutableDictionary dictionary];
  [PtDict setObject:nX forKey:OpenClip_X];
  [PtDict setObject:nY forKey:OpenClip_Y];
  
  [Points addObject:PtDict];
};


void CopyJob::AddXYZPoint(double X,double Y,double Z,NSMutableArray *Points) const
{
  NSNumber *nX = [NSNumber numberWithDouble:X];  // this is how you add an XYZ point if you want to
  NSNumber *nY = [NSNumber numberWithDouble:Y];
  NSNumber *nZ = [NSNumber numberWithDouble:Z];

  NSMutableDictionary *PtDict	= [NSMutableDictionary dictionary];
  [PtDict setObject:nX forKey:OpenClip_X];
  [PtDict setObject:nY forKey:OpenClip_Y];
  [PtDict setObject:nZ forKey:OpenClip_Z];
  
  [Points addObject:PtDict];
};


void CopyJob::AddPointsToObject(NSMutableArray *Points,NSMutableDictionary *TheObj) const
{
  [TheObj setObject:Points forKey:OpenClip_Points];
};


void CopyJob::SetObjectZ(double Z,NSMutableDictionary *TheObj) const
{
  NSNumber *nZ = [NSNumber numberWithDouble:Z];
  [TheObj setObject:nZ forKey:OpenClip_Z];
};


NSMutableDictionary* CopyJob::RGBADictionary(RGBColor TheColor,double Alpha) const
{
  float R = (float)TheColor.red / 65535.0;
  float G = (float)TheColor.green / 65535.0;
  float B = (float)TheColor.blue / 65535.0;
  
  NSNumber *nR = [NSNumber numberWithFloat:R];
  NSNumber *nG = [NSNumber numberWithFloat:G];
  NSNumber *nB = [NSNumber numberWithFloat:B];
  NSNumber *nA = [NSNumber numberWithFloat:Alpha];
  
  NSMutableDictionary *RGBA	= [NSMutableDictionary dictionary];
  [RGBA setObject:nR forKey:OpenClip_Red];
  [RGBA setObject:nG forKey:OpenClip_Green];
  [RGBA setObject:nB forKey:OpenClip_Blue];
  [RGBA setObject:nA forKey:OpenClip_Alpha];
  
  return RGBA;
};


void CopyJob::SetPenColor(RGBColor PenColor,double PenAlpha,NSMutableDictionary *TheObj) const
{
  NSMutableDictionary *TheColor = RGBADictionary(PenColor,PenAlpha);
  [TheObj setObject:TheColor forKey:OpenClip_PenColor];
};


void CopyJob::SetFillColor(const PDObject &TheObj,NSMutableDictionary *ThisObj) const
{
  CPatRecordRef TheFillPattern = TheObj.GetFillPattern();

  if (TheFillPattern != kNonePat()) {
    RGBColor FillColor;
    
    if (TheFillPattern == kWhitePat())
      FillColor = kWhiteRGB;
    else
      FillColor = *TheObj.GetFillColor();
      
    double FillAlpha = TheObj.GetFillAlpha();
    
    NSMutableDictionary *TheColor = RGBADictionary(FillColor,FillAlpha);
    [ThisObj setObject:TheColor forKey:OpenClip_FillColor];
  };
  
};

void CopyJob::SetLineDash(const PDObject &TheObj,NSMutableDictionary *ThisObj) const
{
  if (TheObj.Dashed()) {
    TDashedLine DL = *TheObj.GetDash();
    
    NSMutableArray *Dashes = [NSMutableArray array];
    
    long LastN = DL.LastIndex();
    
    for(long N = 0; (N <= LastN); N++) {
      double DashLength = DL.Dash(N);
      NSNumber *nDashLength = [NSNumber numberWithFloat:DashLength];
      [Dashes addObject:nDashLength];
    };
    
    [ThisObj setObject:Dashes forKey:OpenClip_LineDash];
  }
};


void CopyJob::SetPenSize(float TheSize,NSMutableDictionary *TheObj) const
{
  NSNumber *nTheSize = [NSNumber numberWithFloat:TheSize];
  [TheObj setObject:nTheSize forKey:OpenClip_PenSize];
};


void CopyJob::SetRotationAngle(double TheAngle,NSMutableDictionary *TheObj) const
{
  NSNumber *nTheAngle = [NSNumber numberWithDouble:TheAngle];
  [TheObj setObject:nTheAngle forKey:OpenClip_RotationAngle];
};


void CopyJob::SetStartAngle(double TheAngle,NSMutableDictionary *TheObj) const
{
  NSNumber *nTheAngle = [NSNumber numberWithDouble:TheAngle];
  [TheObj setObject:nTheAngle forKey:OpenClip_StartAngle];
};


void CopyJob::SetArcAngle(double TheAngle,NSMutableDictionary *TheObj) const
{
  NSNumber *nTheAngle = [NSNumber numberWithDouble:TheAngle];
  [TheObj setObject:nTheAngle forKey:OpenClip_ArcAngle];
};


void CopyJob::SetCornerRadius(double TheRadius,NSMutableDictionary *TheObj) const
{
  NSNumber *nTheRadius = [NSNumber numberWithDouble:TheRadius];
  [TheObj setObject:nTheRadius forKey:OpenClip_CornerRadius];
};


void CopyJob::SetClosed(bool IsClosed,NSMutableDictionary *TheObj) const
{
  NSNumber *nIsClosed = [NSNumber numberWithBool:IsClosed];
  [TheObj setObject:nIsClosed forKey:OpenClip_Closed];
};


bool CopyJob::GetArrowObj(const CLineEndType &TheArrow,RGBColor ThePenColor,double PenAlpha,
                          float ThePenSize,PDObject &ThisObj) const
{
	bool GotArrow = false;  // make it bulletproof in case a new type of arrow gets invented
  
  CDPoint ThePt = TheArrow.Location();
	short TheType = TheArrow.EndType();
	short	Misc    = TheArrow.ExtraInfo();
  
  bool IsFilled = BTst(Misc,cEndFilled);
  bool IsThick  = BTst(Misc,cEndThick);
  bool IsCurved = BTst(Misc,cEndCurved);
  
  CDPoint ZeroPt(0,mPgSzY);
  ThePt.y += mPgSzY;
  double Y0 = mPgSzY;
  
  CCMColorRef ThePenColorRef = new CCMColor(ThePenColor);
  
  switch (TheType) {
    case arrowType  : {
                        ThePt.x = -abs(ThePt.x);  // all this dimensional silliness is to put
                        ThePt.y = abs(ThePt.y);   // the arrow at the bottom left of the page
                                                  // so that when it is corrected by CorrectPoint
                        if (IsCurved) {           // it will end up at the drawing origin
                          CDPoint A(ThePt);
                          CDPoint B(A.x * 0.77,Y0 + (A.y - Y0) * 0.46);
                          CDPoint C(A.x * 0.34,Y0 + (A.y - Y0) * 0.12);
                          CDPoint D(ZeroPt);
                          CDPoint E(C.x,Y0 - (A.y - Y0) * 0.12);
                          CDPoint F(B.x,Y0 - (A.y - Y0) * 0.46);
                          CDPoint G(A.x,Y0 - (A.y - Y0));
                          CDPoint H(A.x * 0.92,Y0 - (A.y - Y0) * 0.58);
                          CDPoint I(A.x * 0.90,Y0 - (A.y - Y0) * 0.35);
                          CDPoint J(I.x,D.y);
                          CDPoint K(I.x,Y0 + (A.y - Y0) * 0.35);
                          CDPoint L(H.x,Y0 + (A.y - Y0) * 0.58);
                          
                          StartBez( ThisObj,A,A);
                          BezSegment(ThisObj,B,C,D);
                          BezSegment(ThisObj,E,F,G);
                          
                          if (IsFilled) {
                            BezSegment(ThisObj,H,I,J);
                            BezSegment(ThisObj,K,L,A);
                            EndBez(ThisObj,A);
                            ThisObj.SetClosed(true);
                          }
                          else
                            EndBez(ThisObj,G);
                        
                        }
                        else {
                          InitObject( ThisObj,polyObj);
                          CDPoint A(ThePt);
                          CDPoint B(ThePt.x,Y0 - (A.y - Y0));
                          ThisObj.PHandle().Append(A);
                          ThisObj.PHandle().Append(ZeroPt);
                          ThisObj.PHandle().Append(B);
                          
                          if (IsFilled){
                            ThisObj.PHandle().Append(A);
                            ThisObj.SetClosed(true);
                          };
                          
                        }
                        
                        if (IsFilled) {
                          ThisObj.SetFillPattern(kBlackPat());
                          ThisObj.SetFillColor(ThePenColorRef);
                          ThisObj.SetFillAlpha(PenAlpha);
                        }
                        else
                          ThisObj.SetFillPattern(kNonePat());
                        
                        ThisObj.SetPenColor(ThePenColorRef);
                        ThisObj.SetPenPattern(kBlackPat());
                        ThisObj.SetPenAlpha(PenAlpha);
                        GotArrow = true;
                      }
                      break;
    case circleType : {
                        InitObject( ThisObj,ovalObj);
                        double Rad = LineLength(ZeroPt,ThePt);
                        CDRect TheRect(ZeroPt.x - Rad,ZeroPt.y - Rad,ZeroPt.x + Rad,ZeroPt.y + Rad);
                        ThisObj.SetDataRect(TheRect);
                        
                        ThisObj.SetPenColor(ThePenColorRef);
                        ThisObj.SetPenPattern(kBlackPat());
                        ThisObj.SetPenAlpha(PenAlpha);
                        
                        if (IsFilled) {
                          ThisObj.SetFillPattern(kBlackPat());
                          ThisObj.SetFillColor(ThePenColorRef);
                          ThisObj.SetFillAlpha(PenAlpha);
                        }
                        else
                          ThisObj.SetFillPattern(kNonePat());
                        
                        GotArrow = true;
                      }
                      break;
    case slashType  : {
                        InitObject( ThisObj,lineObj);
                        CDPoint A(-ThePt.x,ThePt.y);
                        CDPoint B(ThePt.x,Y0 - (A.y - Y0));
                        SetLinePts(A,B, ThisObj);
                        
                        ThisObj.SetPenColor(ThePenColorRef);
                        ThisObj.SetPenPattern(kBlackPat());
                        ThisObj.SetPenAlpha(PenAlpha);
                        
                        if (IsThick)
                          ThisObj.SetPenSize(ThePenSize * 2);
                        else
                          ThisObj.SetPenSize(ThePenSize);
                        
                        GotArrow = true;
                      }
                      break;
  };
  
  return GotArrow;
};


void CopyJob::SetArrowsBothEnds(const CLineEndType &TheArrow,RGBColor ThePenColor,double PenAlpha,
                                float ThePenSize,NSMutableDictionary *TheObj) const
{
  PDObject ThisObj;
  
  if (GetArrowObj(TheArrow,ThePenColor,PenAlpha,ThePenSize, ThisObj)) {
    NSMutableArray *TheNSArrowList = [NSMutableArray array];
    AddObjectToList(ThisObj,TheNSArrowList);
    [TheObj setObject:TheNSArrowList forKey:OpenClip_Arrows];
  };

};


void CopyJob::SetArrowAtStart(const CLineEndType &TheArrow,RGBColor ThePenColor,double PenAlpha,
                              float ThePenSize,NSMutableDictionary *TheObj) const
{
  PDObject ThisObj;
  
  if (GetArrowObj(TheArrow,ThePenColor,PenAlpha,ThePenSize, ThisObj)) {
    NSMutableArray *TheNSArrowList = [NSMutableArray array];
    AddObjectToList(ThisObj,TheNSArrowList);
    [TheObj setObject:TheNSArrowList forKey:OpenClip_ArrowAtStart];
  };

};


void CopyJob::SetArrowAtEnd(const CLineEndType &TheArrow,RGBColor ThePenColor,double PenAlpha,
                            float ThePenSize,NSMutableDictionary *TheObj) const
{
  PDObject ThisObj;
  
  if (GetArrowObj(TheArrow,ThePenColor,PenAlpha,ThePenSize, ThisObj)) {
    NSMutableArray *TheNSArrowList = [NSMutableArray array];
    AddObjectToList(ThisObj,TheNSArrowList);
    [TheObj setObject:TheNSArrowList forKey:OpenClip_ArrowAtEnd];
  };

};


void CopyJob::SetArrows(const PDObject &ThisObj,RGBColor PenColor,double Alpha,float PenSize,NSMutableDictionary *TheObj) const
{
  if (ThisObj.ArrowAtStart() && ThisObj.ArrowAtEnd()) {
    CLineEndType Arrow1 = *ThisObj.GetArrow1();
    CLineEndType Arrow2 = *ThisObj.GetArrow2();
    
    if (EqualArrow(Arrow1,Arrow2)) {
      SetArrowsBothEnds(Arrow1,PenColor,Alpha,PenSize,TheObj);
    }
    else {
      SetArrowAtStart(Arrow1,PenColor,Alpha,PenSize,TheObj);
      SetArrowAtEnd(Arrow2,PenColor,Alpha,PenSize,TheObj);
    };
  
  }
  else if (ThisObj.ArrowAtStart()) {
    CLineEndType Arrow1 = *ThisObj.GetArrow1();
    SetArrowAtStart(Arrow1,PenColor,Alpha,PenSize,TheObj);
  }
  else if (ThisObj.ArrowAtEnd()) {
    CLineEndType Arrow2 = *ThisObj.GetArrow2();
    SetArrowAtEnd(Arrow2,PenColor,Alpha,PenSize,TheObj);
  };

};


void CopyJob::CreatePointObj(const PDObject &TheObj,NSMutableDictionary *ThisObj) const
{
  SetObjectType(OpenClip_PointObj,ThisObj);
  
  RGBColor ThePenColor = *TheObj.GetPenColor();
  double PenAlpha = TheObj.GetPenAlpha();
  SetPenColor(ThePenColor,PenAlpha,ThisObj);
  
  HandleZHeight(TheObj,ThisObj);
  
  CDPoint A = TheObj.DataPoint();
  A = CorrectPoint(A);
  
  NSMutableArray *Points = [NSMutableArray array];
  AddXYPoint(A.x,A.y,Points);
  AddPointsToObject(Points,ThisObj);
};


void CopyJob::CreateLineObj(const PDObject &TheObj,NSMutableDictionary *ThisObj) const
{
  SetObjectType(OpenClip_LineObj,ThisObj);
  
  RGBColor ThePenColor = *TheObj.GetPenColor();
  double PenAlpha = TheObj.GetPenAlpha();
  SetPenColor(ThePenColor,PenAlpha,ThisObj);
  
  float ThePenSize = TheObj.GetPenSize();
  SetPenSize(ThePenSize,ThisObj);
  SetLineDash(TheObj,ThisObj);
  SetArrows(TheObj,ThePenColor,PenAlpha,ThePenSize,ThisObj);
  HandleZHeight(TheObj,ThisObj);
  
  CDPoint A;
  CDPoint B;
  GetLinePts(TheObj, A, B);
  
  A = CorrectPoint(A);
  B = CorrectPoint(B);
  
  NSMutableArray *Points = [NSMutableArray array];
  AddXYPoint(A.x,A.y,Points);
  AddXYPoint(B.x,B.y,Points);
  AddPointsToObject(Points,ThisObj);
};


void CopyJob::CreateRectangleObj(const PDObject &TheObj,NSMutableDictionary *ThisObj) const
{
  SetObjectType(OpenClip_RectObj,ThisObj);
  
  RGBColor ThePenColor = *TheObj.GetPenColor();
  double PenAlpha = TheObj.GetPenAlpha();
  SetPenColor(ThePenColor,PenAlpha,ThisObj);
  
  float ThePenSize = TheObj.GetPenSize();
  SetPenSize(ThePenSize,ThisObj);
  SetLineDash(TheObj,ThisObj);
  HandleZHeight(TheObj,ThisObj);
  SetFillColor(TheObj,ThisObj);
  
  CDRect TheRect;
  
  switch (TheObj.WhatObject()) {
    case rectObj   :
    case rrectObj  : TheRect = TheObj.DataRect();
                     break;
    case rtRectObj : TheRect = TheObj.UnRtRect();
                     break;
  };
  
  CDPoint TL = TheRect.TopLeft();
  CDPoint BR = TheRect.BottomRight();
  
  TL = CorrectPoint(TL);
  BR = CorrectPoint(BR);
  
  NSMutableArray *Points = [NSMutableArray array];
  AddXYPoint(TL.x,TL.y,Points);
  AddXYPoint(BR.x,BR.y,Points);
  AddPointsToObject(Points,ThisObj);
  
  if (TheObj.WhatObject() == rtRectObj)
    SetRotationAngle(Radians2Degrees(-TheObj.RotationAngle()),ThisObj);
  
  if (TheObj.WhatObject() == rrectObj) {
    double TheRad = TheObj.OvalRad() / 2.0; // OvalRad is really a diameter
    SetCornerRadius(TheRad,ThisObj);
  };
  
};


void CopyJob::CreateEllipseObj(const PDObject &TheObj,NSMutableDictionary *ThisObj) const
{
  SetObjectType(OpenClip_EllipseObj,ThisObj);
  
  RGBColor ThePenColor = *TheObj.GetPenColor();
  double PenAlpha = TheObj.GetPenAlpha();
  SetPenColor(ThePenColor,PenAlpha,ThisObj);
  
  float ThePenSize = TheObj.GetPenSize();
  SetPenSize(ThePenSize,ThisObj);
  SetLineDash(TheObj,ThisObj);
  
  HandleZHeight(TheObj,ThisObj);
  SetFillColor(TheObj,ThisObj);
  
  CDRect TheRect;
  
  switch (TheObj.WhatObject()) {
    case ovalObj   : TheRect = TheObj.DataRect();
                     break;
    case rOvalObj  :
    case arcObj    :
    case rtArcObj  : TheRect = TheObj.UnRtRect();
                     break;
  };
  
  CDPoint TL = TheRect.TopLeft();
  CDPoint BR = TheRect.BottomRight();
  
  TL = CorrectPoint(TL);
  BR = CorrectPoint(BR);
  
  NSMutableArray *Points = [NSMutableArray array];
  AddXYPoint(TL.x,TL.y,Points);
  AddXYPoint(BR.x,BR.y,Points);
  AddPointsToObject(Points,ThisObj);
  
  switch (TheObj.WhatObject()) {
    case rOvalObj  :
    case rtArcObj  : SetRotationAngle(Radians2Degrees(-TheObj.RotationAngle()),ThisObj);
                     break;
  };

  switch (TheObj.WhatObject()) {  // PowerCADD has very non-standard angles
    case arcObj    :
    case rtArcObj  : double TheStartAngle = 90.0 - Radians2Degrees(TheObj.StartAngle());
                     SetStartAngle(TheStartAngle,ThisObj);
                     
                     double TheArcAngle = Radians2Degrees(-TheObj.ArcAngle());
                     SetArcAngle(TheArcAngle,ThisObj);
                     
                     SetArrows(TheObj,ThePenColor,PenAlpha,ThePenSize,ThisObj);
                     break;
  };
	
};


void CopyJob::CreatePolygonObj(const PDObject &TheObj,NSMutableDictionary *ThisObj) const
{
  SetObjectType(OpenClip_PolygonObj,ThisObj);
  
  RGBColor ThePenColor = *TheObj.GetPenColor();
  double PenAlpha = TheObj.GetPenAlpha();
  SetPenColor(ThePenColor,PenAlpha,ThisObj);
  
  float ThePenSize = TheObj.GetPenSize();
  SetPenSize(ThePenSize,ThisObj);
  
  SetLineDash(TheObj,ThisObj);
  
  HandleZHeight(TheObj,ThisObj);
  SetFillColor(TheObj,ThisObj);
  SetClosed(TheObj.Closed(),ThisObj);
  
  if (!TheObj.Closed())
    SetArrows(TheObj,ThePenColor,PenAlpha,ThePenSize,ThisObj);
  
  NSMutableArray *Points = [NSMutableArray array];
  
  CFPolygonRef ThePoly = TheObj.PHandle();
  long LastN = ThePoly.LastIndex();
  
  for(long N = 0; (N <= LastN); N++) {
    CDPoint A = ThePoly.PolyPoints(N);
    A = CorrectPoint(A);
    AddXYPoint(A.x,A.y,Points);
  };
  
  AddPointsToObject(Points,ThisObj);
};


void CopyJob::AddBSplineAsAdvanced(const PDObject &TheObj,NSMutableDictionary *MainObj) const
{
  NSMutableDictionary *ThisObj = [NSMutableDictionary dictionary];

  SetObjectType(OpenClip_BSplineObj,ThisObj);
  
  // add bspline order ?
  
  RGBColor ThePenColor = *TheObj.GetPenColor();
  double PenAlpha = TheObj.GetPenAlpha();
  SetPenColor(ThePenColor,PenAlpha,ThisObj);
  
  float ThePenSize = TheObj.GetPenSize();
  SetPenSize(ThePenSize,ThisObj);
  SetLineDash(TheObj,ThisObj);
  HandleZHeight(TheObj,ThisObj);
  SetFillColor(TheObj,ThisObj);
  SetClosed(TheObj.Closed(),ThisObj);
  
  if (!TheObj.Closed())
    SetArrows(TheObj,ThePenColor,PenAlpha,ThePenSize,ThisObj);
  
  NSMutableArray *Points = [NSMutableArray array];
  
  CFPolygonRef ThePoly = TheObj.PHandle();
  long LastN = ThePoly.LastIndex();
  
  for(long N = 0; (N <= LastN); N++) {
    CDPoint A = ThePoly.PolyPoints(N);
    A = CorrectPoint(A);
    AddXYPoint(A.x,A.y,Points);
  };
  
  AddPointsToObject(Points,ThisObj);
  
  NSMutableArray *TheAdvancedList	= [NSMutableArray array];
  [TheAdvancedList addObject:ThisObj]; 
  [MainObj setObject:TheAdvancedList forKey:OpenClip_Advanced];
};


void CopyJob::CreateBezierObj(const PDObject &TheObj,NSMutableDictionary *ThisObj) const
{
  SetObjectType(OpenClip_BezierObj,ThisObj);
  
  RGBColor ThePenColor = *TheObj.GetPenColor();
  double PenAlpha = TheObj.GetPenAlpha();
  SetPenColor(ThePenColor,PenAlpha,ThisObj);
  
  float ThePenSize = TheObj.GetPenSize();
  SetPenSize(ThePenSize,ThisObj);
  SetLineDash(TheObj,ThisObj);
  HandleZHeight(TheObj,ThisObj);
  SetFillColor(TheObj,ThisObj);
  SetClosed(TheObj.Closed(),ThisObj);
  
  if (!TheObj.Closed())
    SetArrows(TheObj,ThePenColor,PenAlpha,ThePenSize,ThisObj);
  
  NSMutableArray *Points = [NSMutableArray array];
  
  CFPolygonRef ThePoly = TheObj.PHandle();
  long LastN = ThePoly.LastIndex();
  
  //for(long N = 0; (N <= LastN); N++) {
  for(long N = 1; (N < LastN); N++) {  // skip first and last points
    CDPoint A = ThePoly.PolyPoints(N); // because PowerCADD has a dummy
    A = CorrectPoint(A);               // point at the start and end
    AddXYPoint(A.x,A.y,Points);        // for control handles
  };
  
  AddPointsToObject(Points,ThisObj);
};


void CopyJob::CreateGroupObj(const PDObject &TheObj,NSMutableDictionary *ThisObj) const
{
  SetObjectType(OpenClip_GroupObj,ThisObj);
  
  NSMutableArray *TheNSGroupList = [NSMutableArray array];

  ListRecord GrpList = TheObj.GroupList();
  CheckList(GrpList,false,TheNSGroupList);
  
  if ([TheNSGroupList count] > 0) {
    //add drawing objects to the object
    [ThisObj setObject:TheNSGroupList forKey:OpenClip_DrawingObjects];
  };

};


void CopyJob::AddObjectToList(const PDObject &TheObj,NSMutableArray *TheNSList) const
{
  switch (TheObj.WhatObject()) {
    case pointObj   : {
                        NSMutableDictionary *ThisObj = [NSMutableDictionary dictionary];
                        CreatePointObj(TheObj,ThisObj);
                        
                        [TheNSList addObject:ThisObj]; 
                      }
                      break;
    case lineObj    : {
                        NSMutableDictionary *ThisObj = [NSMutableDictionary dictionary];
                        CreateLineObj(TheObj,ThisObj);
                        
                        [TheNSList addObject:ThisObj]; 
                      }
                      break;
    case arcObj     :
    case rtArcObj   : 
    case ovalObj    : 
    case rOvalObj   : {
                        NSMutableDictionary *ThisObj = [NSMutableDictionary dictionary];
                        CreateEllipseObj(TheObj,ThisObj);
                        
                        [TheNSList addObject:ThisObj]; 
                      }
                      break;
    case rectObj    : 
    case rtRectObj  :
    case rrectObj   : {
                        NSMutableDictionary *ThisObj = [NSMutableDictionary dictionary];
                        CreateRectangleObj(TheObj,ThisObj);
                        
                        [TheNSList addObject:ThisObj]; 
                      }
                      break;
    case polyObj    : {
                        NSMutableDictionary *ThisObj = [NSMutableDictionary dictionary];
                        CreatePolygonObj(TheObj,ThisObj);
                        
                        [TheNSList addObject:ThisObj];
                      };
                      break;
    case smoothObj  : {
                        NSMutableDictionary *ThisObj = [NSMutableDictionary dictionary];
                        
                        PDObject TheBez;
                        BSplineToBezier(TheObj, TheBez);
                        CreateBezierObj(TheBez,ThisObj);
                        RecoverObject( TheBez);
                        
                        AddBSplineAsAdvanced(TheObj,ThisObj);
                        
                        [TheNSList addObject:ThisObj]; 
                      }
                      break;
    case bezObj     : {
                        NSMutableDictionary *ThisObj = [NSMutableDictionary dictionary];
                        CreateBezierObj(TheObj,ThisObj);
                        
                        [TheNSList addObject:ThisObj]; 
                      }
                      break;
    case grpObj     :
    case xtnlGrpObj : {
                        NSMutableDictionary *ThisObj = [NSMutableDictionary dictionary];
                        CreateGroupObj(TheObj,ThisObj);
                        
                        [TheNSList addObject:ThisObj]; 
                      
                      }
                      break;
  };

};


void CopyJob::CheckList(ListRecord TheList,bool InAttachObj,NSMutableArray *TheNSList) const
{
  CPDListIter	Iter(TheList,false,VectorObjects);
	
	ObjSpec			TheObjS;
	
  while (Iter.Next( TheObjS)) {  // we are iterating through a list of objects here
	  PDObject TheObj = TheObjS;
		
		switch (TheObj.WhatObject()) {
      case pointObj   : 
      case lineObj    : 
      case rectObj    :
      case rtRectObj  :
      case rrectObj   :
      case rtArcObj   :
      case arcObj     :
      case ovalObj    :
      case rOvalObj   :
      case polyObj    :
      case smoothObj  :
      case bezObj     : if (InAttachObj) {
                          
                          if (TheObj.GetPenPattern() != kNonePat())
                            AddObjectToList(TheObj,TheNSList);
                        
                        }
                        else
                          AddObjectToList(TheObj,TheNSList);
                        break;
			case grpObj     :
			case xtnlGrpObj : if (mHandleGroups)
                          AddObjectToList(TheObj,TheNSList);
                        else {                       // do something about attach objects
                          ListRecord GrpList = TheObj.GroupList();
                          bool IsAttachObj = (TheObj.SpecialGroupKind() == attachTypeObj);
                          CheckList(GrpList,IsAttachObj,TheNSList);
                        }

                        break;
		};
	};
};


void CopyJob::CheckObjS(ObjSpec TheObjS,NSMutableArray *TheNSList)
{
  PDObject TheObj = TheObjS;

  switch (TheObj.WhatObject()) {
    case pointObj   : 
    case lineObj    : 
    case rectObj    :
    case rtRectObj  :
    case rrectObj   :
    case rtArcObj   :
    case arcObj     :
    case ovalObj    :
    case rOvalObj   :
    case polyObj    :
    case smoothObj  :
    case bezObj     : AddObjectToList(TheObj,TheNSList);
                      break;
    case xtnlGrpObj : 
    case grpObj     : if (mHandleGroups)
                        AddObjectToList(TheObj,TheNSList);
                      else {                       // do something about attach objects
                        ListRecord GrpList = TheObj.GroupList();
                        bool IsAttachObj = (TheObj.SpecialGroupKind() == attachTypeObj);
                        CheckList(GrpList,IsAttachObj,TheNSList);
                      }
                      break;
  };

};


void CopyJob::WriteOutFile(NSMutableDictionary *TheDocument)
{
  NSFileManager *fileManager = [NSFileManager defaultManager];
  NSArray *Paths = NSSearchPathForDirectoriesInDomains(NSApplicationSupportDirectory,NSLocalDomainMask,NO);
  
  if ([Paths count] > 0) {
    //NSString *OpenClipPath = @"/OpenClip";
    //NSLog([OpenClipPath lastPathComponent]);  // OpenClip
    
    //NSString *destinationPath = [[Paths objectAtIndex:0] stringByAppendingPathComponent:[OpenClipPath lastPathComponent]];
    NSString *destinationPath = [[Paths objectAtIndex:0] stringByAppendingPathComponent:@"OpenClip"];
    //NSLog(destinationPath); // /Library/Application Support/OpenClip
    
    if (! [fileManager fileExistsAtPath:destinationPath]) {
      [fileManager createDirectoryAtPath:destinationPath attributes:nil];
      //NSLog(@"Created Directory");
    };
    
    //NSString *TheFilePath = [destinationPath stringByAppendingPathComponent:[OpenClipPath lastPathComponent]];
    NSString *TheFilePath = [destinationPath stringByAppendingPathComponent:@"OpenClip"];
    //NSLog(TheFilePath);  // /Library/Application Support/OpenClip/OpenClip
    
    NSFileHandle *theFileHandle = [NSFileHandle fileHandleForWritingAtPath:TheFilePath];
    
    [TheDocument writeToFile:TheFilePath atomically:YES]; // writeToFile is NSData
    
    //if ([TheDocument writeToFile:TheFilePath atomically:YES])
    //  NSLog(@"Wrote file OK");
    //else
    //  NSLog(@"Did not write file");
    
    [theFileHandle closeFile];  // this is all new stuff for me
  };
};


/*
bool DoPasteboardTool(DWindow *TheWndw,const EventRecord &OrgEvent,const CDPoint &OrgPt)
{

  //long          MDnTicks = OrgEvent.when;
  //UInt32  			MUpTicks;
  CDPoint       Pt1 = OrgPt;  // here I was just trying to find my way around in dealing with the pasteboard
                              // ignore all this, but if you have to go 'digging' this is a place to start
  
  OSStatus err = noErr;
  PasteboardRef theClipboard;
  
  err = PasteboardCreate( kPasteboardClipboard, &theClipboard);
  
  //PasteboardSyncFlags syncFlags;
  ItemCount           itemCount;

  //syncFlags = PasteboardSynchronize( inPasteboard );

  //require_action( syncFlags&kPasteboardModified, PasteboardOutOfSync,err = badPasteboardSyncErr );

  err = PasteboardGetItemCount(theClipboard, &itemCount);
  
  NSLog(@"Pasteboard ItemCount: %d",itemCount);  // = 1
  //require_noerr(err,CantGetPasteboardItemCount);

  for(UInt32 itemIndex = 1; itemIndex <= itemCount; itemIndex++ ) {
    PasteboardItemID    itemID;
    
    err = PasteboardGetItemIdentifier(theClipboard,itemIndex, &itemID );
    
    NSLog(@"Pasteboard itemID: %d",itemID); // = 0

    //require_noerr( err, CantGetPasteboardItemIdentifier );
    
    CFArrayRef          flavorTypeArray;
    err = PasteboardCopyItemFlavors(theClipboard, itemID, &flavorTypeArray );

    //require_noerr( err, CantCopyPasteboardItemFlavors );
    
    CFIndex             flavorCount;
    flavorCount = CFArrayGetCount(flavorTypeArray );
    
    NSLog(@"Pasteboard flavorCount: %d",flavorCount);  // = 4

    for( CFIndex flavorIndex = 0; flavorIndex < flavorCount; flavorIndex++ ) {
      CFStringRef             flavorType;
      //CFDataRef               flavorData;
      //CFIndex                 flavorDataSize;
      char                    flavorText[256];

      flavorType = (CFStringRef)CFArrayGetValueAtIndex( flavorTypeArray,flavorIndex );
      NSLog(@"flavorType: %@",flavorType); // com.engsw.pc8
                                           // dyn.agk8za3bfg6
                                           // com.adobe.pdf
                                           // com.apple.pict

      //if (UTTypeConformsTo(flavorType, CFSTR("public.utf16-plain-text"))) {
      if (UTTypeConformsTo(flavorType, CFSTR("com.apple.pict"))) {
        NSLog(@"conforms");
        
        CFDataRef               flavorData;
        err = PasteboardCopyItemFlavorData(theClipboard,itemID,flavorType, &flavorData);
        //require_noerr( err, CantCopyFlavorData );
        
        CFIndex                 flavorDataSize;
        flavorDataSize = CFDataGetLength( flavorData );
        
        NSLog(@"flavorDataSize: %d",flavorDataSize);  // = 5331
        
        //flavorDataSize = (flavorDataSize<254) ? flavorDataSize : 254;
        
        //const UInt8* DataPtr = CFDataGetBytePtr(flavorData);
        //string s(DataPtr,flavorDataSize);
        string s;

        for( short dataIndex = 0; dataIndex <= flavorDataSize; dataIndex++ ) {
          char byte = *(CFDataGetBytePtr( flavorData ) + dataIndex);
          //flavorText[dataIndex] = byte;
          s += byte;
        }

        //flavorText[flavorDataSize] = '\0';
        //flavorText[flavorDataSize+1] = '\n';
        //TXNSetData( inTXNObject, kTXNTextData, flavorText, flavorDataSize+2, kTXNEndOffset, TXNEndOffset );

        CFRelease (flavorData);
        
        aTextLocRec		Info;
        GetTextLocRec(TheWndw,kHCtrOfText,kVCtrOfText, Info);
        
        CDPoint TextPt = OrgPt;
        //TextPt.Offset(100,100);
        
        PDObject TheTextObj;
        double				TheRad;
        GetTextLocObjString(s,Info,TextPt, TheRad, TheTextObj);
              
        SetTheObject(TheWndw,TheTextObj);
      }

CantCopyFlavorData:
            ;

    }

    CFRelease (flavorTypeArray);

CantCopyPasteboardItemFlavors:
CantGetPasteboardItemIdentifier:
        ;

  }

CantGetPasteboardItemCount:
PasteboardOutOfSync:

    //return err;
  return true;
};
*/


void CopyJob::WriteToPasteboard(NSMutableDictionary *TheDocument)
{
  //OSStatus err = noErr;
  
  PasteboardRef theClipboard;
  
  OSStatus err = PasteboardCreate(kPasteboardClipboard,&theClipboard);
  
  if (mClearPasteboard)
    PasteboardClear(theClipboard);
  
  //PasteboardSyncFlags syncFlags;
  //syncFlags = PasteboardSynchronize( theClipboard ); // 2
  
  NSData *TheData = [NSKeyedArchiver archivedDataWithRootObject:TheDocument];
  err = PasteboardPutItemFlavor(theClipboard,(PasteboardItemID)0,CFSTR("com.openclip.clip"),(CFDataRef)TheData,0);
};


void CopyJob::WriteOutFlatFile(NSMutableDictionary *TheDocument)
{
  NSMutableArray *TheNSList	= [NSMutableArray array];
  
  CDWndwIter	Iter(mTheWndw,true,SelectedVectorObjects);
  ObjSpec		  TheObjS;
    
  while (Iter.Next( TheObjS))
    CheckObjS(TheObjS,TheNSList);
  
  if ([TheNSList count] > 0) {
    [TheDocument setObject:TheNSList forKey:OpenClip_DrawingObjects];
    //long			 StartTicks = TickCount();
    WriteOutFile(TheDocument);
    //long			 Ticks2 = TickCount();
    WriteToPasteboard(TheDocument);
    //long			 Ticks3 = TickCount();
    
    //NSLog(@"File save: %d  Pasteboard save:%d",(Ticks2 - StartTicks),(Ticks3 - Ticks2));
    
    // For 91000 3D points, on a MacPro it took 661 ticks (11 seconds) to save file 
    // and 2877 ticks (47 seconds) to save to the pasteboard,
    // the opposite of what you might expect.  I was timeing it here because I was thinking
    // of taking the file-saving out, but it happens so quickly and is such a convenience,
    // I left it in.
  };

};


void CopyJob::WriteOutLayers(NSMutableDictionary *TheDocument)
{
  NSMutableArray *TheNSLayerList = [NSMutableArray array];
  
  CLayerIter	 LayerIter(mTheWndw,false,RequestedLayers);
  CLayerRecord TheLayer;
  
  while (LayerIter.Next(TheLayer)) {
    NSMutableArray *TheNSList	= [NSMutableArray array];
    
    CPDListIter	ListIter(TheLayer.layerList,false,SelectedVectorObjects);
    ObjSpec			TheObjS;

    while (ListIter.Next( TheObjS))  // iterate through layers
      CheckObjS(TheObjS,TheNSList);
      
    if ([TheNSList count] > 0) {
      NSMutableDictionary *TheNSLayer = [NSMutableDictionary dictionary];
      
      //add drawing objects to the layer
      [TheNSLayer setObject:TheNSList forKey:OpenClip_DrawingObjects];
      
      CStr255 TheName = TheLayer.LayerName();
      SetLayerName(TheName,TheNSLayer);
      
      if (TheLayer.LayerHasCustomColor()) {
        RGBColor TheColor = *TheLayer.LayerColor();
        SetLayerColor(TheColor,TheNSLayer);
      };
      
      if (TheLayer.LayerHasCustomScale()) {
        double TheLayerScale = TheLayer.LayerScale();
        SetLayerScale(TheLayerScale,TheNSLayer);
      };
      
      // add layer to layer list
      [TheNSLayerList addObject:TheNSLayer];
    };
    
  }
  
  if ([TheNSLayerList count] > 0) {
    [TheDocument setObject:TheNSLayerList forKey:OpenClip_DrawingLayers];
    WriteOutFile(TheDocument);
    WriteToPasteboard(TheDocument);
  };

};


void CopyJob::WriteOut(void)
{
	NSAutoreleasePool *thePool = [[NSAutoreleasePool alloc] init];
  
  NSMutableDictionary *TheDocument = [NSMutableDictionary dictionary];
  
  [TheDocument setObject:OpenClip_WildToolsInfo forKey:OpenClip_Creator];
  
  bool AllLayersOn = GetAllLayers();
  
  if (AllLayersOn) {
    mScale = mTheWndw->DefaultDrawingScale();
  }
  else {
    UnitRec TheUnits = mTheWndw->GetUnits();
    //double TheScale = TheUnits.Scale();
    mScale = TheUnits.Scale();
  };
  
  //NSLog(@"mScale %g",mScale);
  
  SetScale(mScale,TheDocument);
  
  //SetGlobalOrigin(12345.0,54321.0,TheDocument);
  // The idea of a global origin is for very large drawings where the drawing is in a
  // state plane coordinate system.  This feature is provided in OpenClip with the tags
  // needed, but we've never actually used it.
  
  //SetInsertionPoint(100.0,200.0,TheDocument);
  // The idea behind an insertion point is that some programs have a 'paste at mouse'
  // capability, and the tags are provided for that purpose, but we've never actually
  // used it.
  
  //NSLog(@"CopyJob WriteOut");
  
  switch (mMode) {
    case kForFlatFile  : WriteOutFlatFile(TheDocument);
                         break;
    case kHandleLayers : WriteOutLayers(TheDocument);
                         break;
  };
  
  [thePool drain];
};

//----------------------------------------------------------------------------------------------